%% Implementation of the Truncated Amplitude Flow algorithm proposed in the paper
%  `` Solving Systems of Random Quadratic Equations via Truncated Amplitude
%  Flow'' by G. Wang, G. B. Giannakis, and Y. C. Eldar.
%  The code below is adapted from implementation of the (Truncated) Wirtinger Flow 
% algorithm implemented by E. Candes, X. Li, M. Soltanolkotabi, and Y. Chen.

function [z, time] = TAF1D(y, x, Amatrix)
tic;

if exist('Params', 'var')           == 0,  Params.n2            = 1;    end
if isfield(Params, 'n1')            == 0,  Params.n1            = length(x); end             % signal dimension
if isfield(Params, 'm')             == 0,  Params.m             = floor(length(y)/length(x) * Params.n1) ;  end     % number of measurements
if isfield(Params, 'cplx_flag')     == 0,  Params.cplx_flag     = 1;    end             % real: cplx_flag = 0;  complex: cplx_flag = 1;
if isfield(Params, 'T')             == 0,  Params.T             = 1000;  end    	% number of iterations
if isfield(Params, 'mu')            == 0,  Params.mu            = 0.6 * (1 - Params.cplx_flag) + .9 * Params.cplx_flag;  end		% step size / learning parameter
if isfield(Params, 'gamma_lb')      == 0,  Params.gamma_lb      = .7;   end	% thresholding of throwing small entries of Az 0.3 is a good one
if isfield(Params, 'npower_iter')   == 0,  Params.npower_iter   = 100;   end		% number of power iterations


%% Initialization
z0      = randn(Params.n1, Params.n2);
z0      = z0 / norm(z0,'fro');    % Initial guess
normest = sqrt(sum(y(:)) / numel(y(:)));    % Estimate norm to scale eigenvector
m       = Params.m;

Arnorm  = sqrt(sum(abs(Amatrix).^2, 2)); % norm of rows of Amatrix
ymag    = sqrt(y);
ynorm   = ymag ./ (Arnorm .* normest);

%% finding the angle of inner prodcut

Anorm      = bsxfun(@rdivide, Amatrix, Arnorm);
[ysort, ~] = sort(ynorm, 'ascend');
 ythresh    = ysort(round(Params.m / (6/5))); % 6/5 the orthogonality-promoting initialization parameter
ind        = (abs(ynorm) >= ythresh);

for tt = 1:Params.npower_iter                   % Truncated power iterations
    z0 = Anorm' * (ind .* (Anorm * z0));
    z0 = z0 /norm(z0, 'fro');
end

z          = normest * z0;                   % Apply scaling
Relerrs = norm(x - exp(-1i * angle(trace(x' * z))) * z, 'fro') / norm(x, 'fro'); % Initial rel. error

for t = 1: Params.T
    
    Az    = Amatrix * z; %A(z);
    ratio = abs(Az) ./ ymag;
    yz    = ratio > 1 / (1 + Params.gamma_lb);
    ang   = Params.cplx_flag * exp(1i * angle(Az)) + (1 - Params.cplx_flag) * sign(Az);
    
    grad  = Amatrix' * (yz .* ymag .* ang - yz .* Az) / m;
    z     = z + Params.mu * grad;
    
    Relerrs = [Relerrs; norm(x - exp(-1i * angle(trace(x' * z))) * z, 'fro') / norm(x,'fro')]; %#ok<AGROW>

end
z = z*exp(1j*angle(z'*x));

time = toc;
end


